/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Netscape Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s): 
 */
#include "nsPGP.h"
#include "nsPGPModule.h"
#include "prmem.h"
#include "mimexpcom.h"
#include "mimecth.h"
#include "msgCore.h"
#include "nsIComponentLoader.h"

static int parse_line (char *, PRInt32, MimeObject *);
static int parse_eof (MimeObject *, PRBool);
static int parse_begin (MimeObject *obj);

static NS_DEFINE_CID(kPGPPlainTextCID, NS_PGP_PLAINTEXT_CONTENT_TYPE_HANDLER_CID);
static NS_DEFINE_CID(kPGPModuleCID, NS_PGP_MODULE_CID); 

/*  This is the object definition. Note: we will set the superclass 
    to NULL and manually set this on the class creation */

MimeDefClass(MimePGP, MimePGPClass, mimePGPClass, NULL);  

extern "C" MimeObjectClass *
MIME_CreateContentTypeHandlerClass(const char *content_type, 
                                   contentTypeHandlerInitStruct *initStruct)
{
  MimeObjectClass *clazz = (MimeObjectClass *)&mimePGPClass;
  /* 
  * Must set the superclass by hand.
  */
  if (!COM_GetmimeMultipartClass())
    return NULL;
  
  clazz->superclass = (MimeObjectClass *)COM_GetmimeMultipartClass();
  initStruct->force_inline_display = PR_TRUE;
  nsCRT::zero(mimePGPClass.content_type, PGP_MAX_CT_SIZE); 
  nsCRT::memcpy(mimePGPClass.content_type, content_type, nsCRT::strlen(content_type));
  return clazz;
}

static int
MimePGPClassInitialize(MimePGPClass *clazz)
{
  MimeObjectClass *oclass = (MimeObjectClass *) clazz;
  PR_ASSERT(!oclass->class_initialized);
  oclass->parse_begin = parse_begin;
  oclass->parse_line  = parse_line;
  oclass->parse_eof   = parse_eof;
  clazz->input = nsnull;
  
  return 0;
}

static int
parse_begin(MimeObject *obj)
{
  MimePGPClass *clazz;
  int status;
  MimeObjectClass *leafClass;
  
  leafClass = ((MimeObjectClass*)COM_GetmimeLeafClass());
  if (!leafClass->class_initialized)
    leafClass->class_initialize(leafClass);
  
  status = leafClass->parse_begin(obj);
  if (status < 0) 
    return status;
  
  if (!obj->output_p) 
    return 0;
  if (!obj->options || !obj->options->write_html_p) 
    return 0;
  
  /* This is a fine place to write out any HTML before the real meat begins. */  
  
  // Initialize the clazz variable...
  clazz = ((MimePGPClass *) obj->clazz);
  
  // Initialize the input variable
  clazz->input = (char *) PR_Malloc(1);
  clazz->input[0] = 0;
  
  return 0;
}

static int
parse_line(char *line, PRInt32 length, MimeObject *obj)
{
  int input_length;
  MimePGPClass *clazz = ((MimePGPClass *) obj->clazz);
  
  /* 
  * This routine gets fed each line of data, one at a time. We just buffer
  * it all up, to be dealt with all at once at the end. 
  */  
  if (!obj->output_p) 
    return 0;
  if (!obj->options || !obj->options->output_fn) 
    return 0;
  
  if (!obj->options->write_html_p) 
  {
    return COM_MimeObject_write(obj, line, length, PR_TRUE);
  }
  
  input_length = nsCRT::strlen(clazz->input);
  clazz->input = (char *) PR_Realloc(clazz->input, input_length + length + 1);
  nsCRT::memcpy(clazz->input + input_length, line, length);
  clazz->input[input_length + length] = 0;
  
  return 0;
}

static int
parse_eof (MimeObject *obj, PRBool abort_p)
{
  int status = 0;
  char* html = NULL;
  MimePGPClass *clazz = ((MimePGPClass *) obj->clazz);
  MimeMultipart *mult = (MimeMultipart *) obj;
  
  if (obj->closed_p) 
    return 0;
  
  /* Run parent method first, to flush out any buffered data. */
  status = ((MimeObjectClass*)COM_GetmimeMultipartClass())->parse_eof(obj, abort_p);
  if (status < 0) 
    return status;
  
  nsCOMPtr<nsIPGPModule> pgp_module;
  
  nsresult res = nsComponentManager::CreateInstance(kPGPModuleCID, 
                  NULL, NS_GET_IID(nsIPGPModule), 
                  (void **) getter_AddRefs(pgp_module)); 
  
  if (!NS_SUCCEEDED(res) || !pgp_module)
    return 0;
  
  void *input = nsnull;
  void *output = nsnull;
  char *output_str = nsnull;
  PRBool pgp_mime = PR_FALSE;
  
  if ((obj->headers->all_headers) && 
    (obj->headers->all_headers_size))
  {
    char *temp_string;
    char *content_type;
    char *mime_version;
    char *mime_type;
    char *mime_version_str = "Mime-Version: 1.0\r\n";
    char *content_type_str = "Content-Type: ";
    char *hdr_start = obj->headers->all_headers;
    char *hdr_end;
    int version_length = nsCRT::strlen(mime_version_str);
    int content_type_len;
    int hdr_length = obj->headers->all_headers_size;
    int total_len;
    
    pgp_module->FindHeader(hdr_start, mime_version_str, &mime_version, 
      &content_type_len, &hdr_end);
    
    pgp_module->FindHeader(hdr_start, content_type_str, &content_type, 
      &content_type_len, &hdr_end);
    
    mime_type = content_type + nsCRT::strlen(content_type_str);
    
    if (content_type_len && mime_version && 
      nsCRT::strncmp(mime_type, "text/plain", nsCRT::strlen("text/plain")))
    {
      temp_string = (char *) PR_Malloc(nsCRT::strlen(clazz->input) + 
                hdr_length + 5);
      
      nsCRT::memcpy(temp_string, mime_version_str, version_length);
      total_len = version_length;
      nsCRT::memcpy(&(temp_string[total_len]), content_type,
        content_type_len);
      total_len += content_type_len;
      nsCRT::memcpy(&(temp_string[total_len]), MSG_LINEBREAK, MSG_LINEBREAK_LEN);
      total_len += MSG_LINEBREAK_LEN;
      nsCRT::memcpy(&(temp_string[total_len]), clazz->input, 
        nsCRT::strlen(clazz->input) + 1);
      
      PR_Free(clazz->input);
      clazz->input = temp_string;
      pgp_mime = PR_TRUE;
    }
  }
  
  pgp_module->CreateMsgBody(clazz->input, &input);
  res = pgp_module->DecryptVerify(0, input, &output);
  pgp_module->GetStringFromMsgBody(output, &output_str);
  
  if (!NS_SUCCEEDED(res) || !output_str)
  {
    pgp_module->FreeMsgBody(input);
    pgp_module->FreeMsgBody(output);
    PR_Free(clazz->input);
    return -1;
  }
  
  MimePartBufferData *part_buffer;
  MimeHeaders *hdrs;
  
  part_buffer = (MimePartBufferData *)COM_MimePartBufferCreate();
  hdrs = (MimeHeaders *)COM_MimeHeaders_new();
  
  PRBool got_header = !pgp_mime;
  PRBool got_cr = PR_FALSE;
  PRBool got_lf = PR_FALSE;
  char line[1024];
  char *line_begin = output_str;
  char *line_end = line_begin;
  
  while (*line_begin)
  {
    while (*line_end && (*line_end != '\n') && (*line_end != '\r'))
      line_end++;
    
    got_cr = PR_FALSE;
    got_lf = PR_FALSE;
    while (1)
    {
      if (*line_end == '\r')
      {
        if (got_cr)
          break;
        else
          got_cr = PR_TRUE;
      }
      else if (*line_end == '\n')
      {
        if (got_lf)
          break;
        else
          got_lf = PR_TRUE;
      }
      else
        break;
      
      line_end++;
    }
    
    nsCRT::memcpy(line, line_begin, line_end - line_begin);
    line[line_end - line_begin] = 0;
    
    if (got_header)
      COM_MimePartBufferWrite ((void *)part_buffer, line, strlen(line));
    else
      COM_MimeHeaders_parse_line(line, nsCRT::strlen(line), (void *)hdrs);
    
    if ((line[0] == '\r') || (line[0] == '\n'))
      got_header = PR_TRUE;
    
    line_begin = line_end;
  }
  
  COM_MimePartBufferWrite ((void *)part_buffer, MSG_LINEBREAK, MSG_LINEBREAK_LEN);
  
  char *ct = COM_MimeHeaders_get ((void *)hdrs, HEADER_CONTENT_TYPE,
    PR_TRUE, PR_FALSE);
  
  MimeObject *body;
  
  if (!nsCRT::strcmp(clazz->input, output_str))
  {
    nsIFactory *factory;
    
    nsComponentManager::FindFactory(kPGPPlainTextCID, &factory);
    nsComponentManager::UnregisterFactory(kPGPPlainTextCID, factory);
    body = (MimeObject *) COM_mime_create("text/plain", (void *)hdrs, 
                            (void *)(obj->options));
    nsComponentManager::RegisterFactory(kPGPPlainTextCID, 
      NS_PGP_PLAINTEXT_CONTENT_TYPE_HANDLER_NAME, 
      NS_PGP_PLAINTEXT_CONTENT_TYPE_HANDLER_CONTRACTID,
      factory,
      PR_TRUE);
    factory->Release();
  }
  else
    body = (MimeObject *) COM_mime_create(ct, (void *)hdrs, 
    (void *)(obj->options));
  
  PR_Free(clazz->input);
  pgp_module->FreeMsgBody(input);
  pgp_module->FreeMsgBody(output);
  PR_FREEIF(ct);
  
  status = ((MimeContainerClass *) obj->clazz)->add_child(obj, body);
  if (status < 0)
  {
    COM_MimePartBufferClose((void *)part_buffer);
    COM_MimeHeaders_free((void *)hdrs);
    return status;
  }
  
  /* Now that we've added this new object to our list of children,
   start its parser going. */
  status = body->clazz->parse_begin(body);
  if (status < 0)
  {
    COM_MimePartBufferClose((void *)part_buffer);
    COM_MimeHeaders_free((void *)hdrs);
    return status;
  }
  
  status = COM_MimePartBufferRead ((void *)part_buffer,
  /* The (int (*) ...) cast is to turn the
  `void' argument into `MimeObject'. */
  ((int (*) (char *, PRInt32, void *))
  body->clazz->parse_buffer),
  (void *)body);
  
  COM_MimePartBufferClose((void *)part_buffer);
  COM_MimeHeaders_free((void *)hdrs);
  
  if (status < 0) return status;
  
  /* Done parsing. */
  status = body->clazz->parse_eof(body, PR_FALSE);
  if (status < 0) return status;
  
  status = body->clazz->parse_end(body, PR_FALSE);
  return status;
}
