/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Netscape Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s): 
 */

#include "nsPGPModule.h"
#include "PGPmoz.h"
#include "prmem.h"
#include "nsCRT.h"

#if defined(_WIN32) && defined(_M_IX86)
#include <windows.h>
#endif /* _WIN32 && _M_IX86 */

struct _PGPvtbl
{
  PGPmozCleanupProc       Cleanup;
  PGPmozEncryptSignProc   EncryptSign;
  PGPmozDecryptVerifyProc DecryptVerify;
  PGPmozFreeMsgBodyProc   FreeMsgBody;
};

typedef struct _PGPmozLocalMsgBody
{
  PRBool          freeLocally;
  PGPmozMsgBody * msgBody;
} PGPmozLocalMsgBody;

void PlatformLoadPlugin(PGPvtbl **vtbl, void **module_object, 
                        void **library_object)
{
  *vtbl = nsnull;
  *module_object = nsnull;
  *library_object = nsnull;
  
#if defined(_WIN32) && defined(_M_IX86)
  // Load the plug-in DLL on Windows platforms
  HINSTANCE hDLL = LoadLibrary("PGPmoz.dll");
  PGPmozInitProc PGPmozInit;
  PGPmozRef mozRef;
  PGPmozError loadError;
  
  if (hDLL)
  {
    PGPmozInit = (PGPmozInitProc) GetProcAddress((HMODULE) hDLL, "PGPmozInit");
    if (PGPmozInit)
    {
      loadError = PGPmozInit(PGPMOZ_VERSION, &mozRef);
      if (loadError == PGPmoz_NoError)
        *module_object = mozRef;
    }
    
    if (!(*module_object))
      FreeLibrary(hDLL);
    else
    {
      *library_object = (void *) hDLL;
      *vtbl = (PGPvtbl *) PR_Malloc(sizeof(PGPvtbl));
      (*vtbl)->Cleanup = (PGPmozCleanupProc) GetProcAddress(hDLL, "PGPmozCleanup");
      (*vtbl)->EncryptSign = (PGPmozEncryptSignProc) GetProcAddress(hDLL, "PGPmozEncryptSign");
      (*vtbl)->DecryptVerify = (PGPmozDecryptVerifyProc) GetProcAddress(hDLL, "PGPmozDecryptVerify");
      (*vtbl)->FreeMsgBody = (PGPmozFreeMsgBodyProc) GetProcAddress(hDLL, "PGPmozFreeMsgBody");
    }
  }
#else
  /* No non-Win32 platform support at this time */
#endif /* _WIN32 && _M_IX86 */
}

/* 
* The following macros actually implement addref, release and 
* query interface for our component. 
*/
NS_IMPL_ISUPPORTS1(nsPGPModule, nsIPGPModule)

/*
* nsPGPModule definitions....
*/

/* 
* Inherited methods for nsPGPModule
*/
nsPGPModule::nsPGPModule()
{
  /* the following macro is used to initialize the ref counting data */
  NS_INIT_REFCNT();
  
  PlatformLoadPlugin(&vtbl, &module_object, &library_object);
}

nsPGPModule::~nsPGPModule()
{
  if (vtbl)
  {
    vtbl->Cleanup((PGPmozRef) module_object);
    PR_Free(vtbl);
  }
  
  if (module_object)
  {
#if defined(_WIN32) && defined(_M_IX86)
    FreeLibrary((HINSTANCE) module_object);
#endif /* _WIN32 && _M_IX86 */
  }
}

nsresult
nsPGPModule::SetEncrypt(PRBool encrypt, int *flags)
{
  nsresult rc = NS_ERROR_FAILURE;
  
  if (flags)
  {
    rc = NS_OK;
    if (encrypt)
      *flags |= PGPmozEncryptSign_Encrypt;
    else
      *flags &= ~PGPmozEncryptSign_Encrypt;
  }
  
  return rc;
}

nsresult
nsPGPModule::SetSign(PRBool sign, int *flags)
{
  nsresult rc = NS_ERROR_FAILURE;
  
  if (flags)
  {
    rc = NS_OK;
    if (sign)
      *flags |= PGPmozEncryptSign_Sign;
    else
      *flags &= ~PGPmozEncryptSign_Sign;
  }
  
  return rc;
}

nsresult
nsPGPModule::SetPGPMIME(PRBool mime, int *flags)
{
  nsresult rc = NS_ERROR_FAILURE;
  
  if (flags)
  {
    rc = NS_OK;
    if (mime)
      *flags |= PGPmozEncryptSign_PGPMIME;
    else
      *flags &= ~PGPmozEncryptSign_PGPMIME;
  }
  
  return rc;
}

nsresult
nsPGPModule::CreateMsgHeader(char *to, char *cc, char *bcc, 
                             void **msg_header)
{
  nsresult rc = NS_ERROR_FAILURE;
  PGPmozMsgHeader *msgHeader;
  
  if (msg_header)
  {
    rc = NS_OK;
    msgHeader = (PGPmozMsgHeader *) PR_Malloc(sizeof(PGPmozMsgHeader));
    
    if (to)
    {
      msgHeader->szTo = (char *) PR_Malloc(nsCRT::strlen(to) + 1);
      PL_strcpy(msgHeader->szTo, to); 
    }
    else
      msgHeader->szTo = nsnull;
    
    if (cc)
    {
      msgHeader->szCc = (char *) PR_Malloc(nsCRT::strlen(cc) + 1);
      PL_strcpy(msgHeader->szCc, cc); 
    }
    else
      msgHeader->szCc = nsnull;
    
    if (bcc)
    {
      msgHeader->szBcc = (char *) PR_Malloc(nsCRT::strlen(bcc) + 1);
      PL_strcpy(msgHeader->szBcc, bcc); 
    }
    else
      msgHeader->szBcc = nsnull;
    
    *msg_header = (void *) msgHeader;
  }
  
  return rc;
}

nsresult
nsPGPModule::CreateMsgBody(char *body, void **msg_body)
{
  nsresult rc = NS_ERROR_FAILURE;
  PGPmozLocalMsgBody *localMsgBody;
  
  if (msg_body)
  {
    rc = NS_OK;
    localMsgBody = (PGPmozLocalMsgBody *) PR_Malloc(sizeof(PGPmozLocalMsgBody));
    
    if (body)
    {
      localMsgBody->msgBody = (PGPmozMsgBody *) PR_Malloc(sizeof(PGPmozMsgBody));
      localMsgBody->msgBody->szBody = (char *) PR_Malloc(nsCRT::strlen(body) + 1);
      PL_strcpy(localMsgBody->msgBody->szBody, body); 
      localMsgBody->freeLocally = PR_TRUE;
    }
    else
    {
      localMsgBody->msgBody = nsnull;
      localMsgBody->freeLocally = PR_FALSE;
    }
    
    *msg_body = (void *) localMsgBody;
  }
  
  return rc;
}

nsresult
nsPGPModule::GetStringFromMsgBody(void *msg_body, char **string)
{
  nsresult rc = NS_ERROR_FAILURE;
  
  if (msg_body && string)
  {
    PGPmozLocalMsgBody *localMsgBody = (PGPmozLocalMsgBody *) msg_body;
    
    if (localMsgBody->msgBody)
      *string = localMsgBody->msgBody->szBody;
    else
      *string = nsnull;
    
    rc = NS_OK;
  }
  
  return rc;
}

nsresult
nsPGPModule::EncryptSign(void *user_value, int flags, void *msg_header,
                         void *orig_body, void **new_body)
{
  nsresult rc = NS_ERROR_FAILURE;
  
  if (vtbl && msg_header && orig_body && new_body)
  {
    PGPmozError error;
    PGPmozLocalMsgBody *origMsgBody;
    PGPmozLocalMsgBody *newMsgBody;
    
    origMsgBody = (PGPmozLocalMsgBody *) orig_body;
    CreateMsgBody(nsnull, (void **) &newMsgBody);
    
    error = vtbl->EncryptSign((PGPmozRef) module_object, user_value, 
              (PGPmozEncryptSignFlags) flags, 
              (PGPmozMsgHeader *) msg_header, 
              origMsgBody->msgBody, &(newMsgBody->msgBody));
    
    if (error == PGPmoz_NoError)
    {
      rc = NS_OK;
      *new_body = (void *) newMsgBody;
    }
  }
  
  return rc;
}

nsresult 
nsPGPModule::DecryptVerify(void *user_value, void *orig_body, 
                           void **new_body)
{
  nsresult rc = NS_ERROR_FAILURE;
  
  if (vtbl && orig_body && new_body)
  {
    PGPmozError error;
    PGPmozLocalMsgBody *origMsgBody;
    PGPmozLocalMsgBody *newMsgBody;
    
    origMsgBody = (PGPmozLocalMsgBody *) orig_body;
    CreateMsgBody(nsnull, (void **) &newMsgBody);
    
    error = vtbl->DecryptVerify((PGPmozRef) module_object, user_value, 
              origMsgBody->msgBody, &(newMsgBody->msgBody));
    
    if (error == PGPmoz_NoError)
    {
      rc = NS_OK;
      *new_body = (void *) newMsgBody;
    }
  }
  
  return rc;
}

nsresult 
nsPGPModule::FreeMsgHeader(void *msg_header)
{
  nsresult rc = NS_ERROR_FAILURE;
  PGPmozMsgHeader *msgHeader = (PGPmozMsgHeader *) msg_header;
  
  if (msgHeader)
  {
    if (msgHeader->szTo)
      PR_Free(msgHeader->szTo);
    
    if (msgHeader->szCc)
      PR_Free(msgHeader->szCc);
    
    if (msgHeader->szBcc)
      PR_Free(msgHeader->szBcc);
    
    PR_Free(msgHeader);
    rc = NS_OK;
  }
  
  return rc;
}

nsresult 
nsPGPModule::FreeMsgBody(void *msg_body)
{
  nsresult rc = NS_ERROR_FAILURE;
  PGPmozLocalMsgBody *localMsgBody = (PGPmozLocalMsgBody *) msg_body;
  
  if (localMsgBody)
  {
    if (localMsgBody->freeLocally)
    {
      if (localMsgBody->msgBody)
        PR_Free(localMsgBody->msgBody->szBody);
      
      PR_Free(localMsgBody->msgBody);
      rc = NS_OK;
    }
    else if (vtbl)
    {
      PGPmozError error;
      
      error = vtbl->FreeMsgBody((PGPmozRef) module_object,
                (PGPmozMsgBody *) localMsgBody->msgBody);
      
      if (error == PGPmoz_NoError)
        rc = NS_OK;
    }
    
    PR_Free(localMsgBody);
  }
  
  return rc;
}

nsresult 
nsPGPModule::FindHeader(char *header_buffer, char *header_str,
                        char **header_start, int *header_length, 
                        char **buffer_end)
{
  char *line_end;
  PRBool end_of_buffer;
  PRBool got_cr;
  PRBool got_lf;
  
  *buffer_end = header_buffer;
  line_end = header_buffer;
  end_of_buffer = PR_FALSE;
  *header_start = 0;
  *header_length = 0;
  
  while (**buffer_end && !end_of_buffer)
  {
    while (*line_end && (*line_end != '\r') && (*line_end != '\n'))
      line_end++;
    
    got_cr = PR_FALSE;
    got_lf = PR_FALSE;
    while (1)
    {
      if (*line_end == '\r')
      {
        if (got_cr)
        {
          end_of_buffer = PR_TRUE;
          break;
        }
        else
          got_cr = PR_TRUE;
      }
      else if (*line_end == '\n')
      {
        if (got_lf)
        {
          end_of_buffer = PR_TRUE;
          break;
        }
        else
          got_lf = PR_TRUE;
      }
      else
        break;
      
      line_end++;
    }
    
    if (end_of_buffer)
      line_end--;
    
    if (!*header_start && 
      !nsCRT::strncmp(*buffer_end, header_str, nsCRT::strlen(header_str)))
    {
      *header_start = *buffer_end;
    }
    
    if (*header_start && !*header_length)
    {
      if (*(line_end - 1 - NS_LINEBREAK_LEN) != ';')
        *header_length = line_end - *header_start;
    }
    
    *buffer_end = line_end;
  }
  
  *buffer_end = line_end;
  
  return NS_OK;
}

